@@ -17,21 +17,25 @@
 from zc.buildout.rmtree import rmtree
 import zc.buildout.easy_install
 
-try:
-    from hashlib import md5
-except ImportError:
-    from md5 import md5
+from functools import partial
 
 try:
-    from UserDict import DictMixin
+    from hashlib import md5 as md5_original
 except ImportError:
-    from collections import MutableMapping as DictMixin
+    from md5 import md5 as md5_original
+
+try:
+    from collections.abc import MutableMapping as DictMixin
+except ImportError:
+    from UserDict import DictMixin
 
 import zc.buildout.configparser
 import copy
 import datetime
 import distutils.errors
 import glob
+import importlib
+import inspect
 import itertools
 import logging
 import os
@@ -50,6 +54,12 @@ if PY3:
 else:
     text_type = unicode
 
+try:
+    hashed = md5_original(b'test')
+    md5 = md5_original
+except ValueError:
+    md5 = partial(md5_original, usedforsecurity=False)
+
 
 def command(method):
     method.buildout_command = True
@@ -108,6 +118,8 @@ class SectionKey(object):
         self.value = sectionkey.value
         if sectionkey.history[-1].operation not in ['ADD', 'REMOVE']:
             self.addToHistory("OVERRIDE", sectionkey.value, sectionkey.source)
+        else:
+            self.history = copy.deepcopy(sectionkey.history)
 
     def setDirectory(self, value):
         self.value = value
@@ -249,15 +261,11 @@ def _print_annotate(data, verbose, chosen_sections, basedir):
 
 
 def _unannotate_section(section):
-    for key in section:
-        section[key] = section[key].value
-    return section
+    return {key: entry.value for key, entry in section.items()}
 
 
 def _unannotate(data):
-    for key in data:
-        data[key] = _unannotate_section(data[key])
-    return data
+    return {key: _unannotate_section(section) for key, section in data.items()}
 
 
 def _format_picked_versions(picked_versions, required_by):
@@ -302,13 +310,19 @@ _buildout_default_options = _annotate_section({
     }, 'DEFAULT_VALUE')
 
 
+def _get_user_config():
+    buildout_home = os.path.join(os.path.expanduser('~'), '.buildout')
+    buildout_home = os.environ.get('BUILDOUT_HOME', buildout_home)
+    return os.path.join(buildout_home, 'default.cfg')
+
+
 @commands
 class Buildout(DictMixin):
 
     COMMANDS = set()
 
     def __init__(self, config_file, cloptions,
-                 user_defaults=True,
+                 use_user_defaults=True,
                  command=None, args=()):
 
         __doing__ = 'Initializing.'
@@ -350,27 +364,45 @@ class Buildout(DictMixin):
         override = copy.deepcopy(cloptions.get('buildout', {}))
 
         # load user defaults, which override defaults
-        if user_defaults:
-            if os.environ.get('BUILDOUT_HOME'):
-                buildout_home = os.environ['BUILDOUT_HOME']
-            else:
-                buildout_home = os.path.join(
-                    os.path.expanduser('~'), '.buildout')
-            user_config = os.path.join(buildout_home, 'default.cfg')
-            if os.path.exists(user_config):
-                data_buildout_copy = copy.deepcopy(data['buildout'])
-                _update(data, _open(os.path.dirname(user_config), user_config,
-                                    [], data_buildout_copy, override,
-                                    set()))
+        user_config = _get_user_config()
+        if use_user_defaults and os.path.exists(user_config):
+            download_options = data['buildout']
+            user_defaults, _ = _open(
+                os.path.dirname(user_config),
+                user_config, [], download_options,
+                override, set(), {}
+            )
+            for_download_options = _update(data, user_defaults)
+        else:
+            user_defaults = {}
+            for_download_options = copy.deepcopy(data)
 
         # load configuration files
         if config_file:
-            data_buildout_copy = copy.deepcopy(data['buildout'])
-            _update(data, _open(os.path.dirname(config_file), config_file, [],
-                                data_buildout_copy, override, set()))
+            download_options = for_download_options['buildout']
+            cfg_data, _ = _open(
+                os.path.dirname(config_file),
+                config_file, [], download_options,
+                override, set(), user_defaults
+            )
+            data = _update(data, cfg_data)
+
+        # extends from command-line
+        if 'buildout' in cloptions:
+            cl_extends = cloptions['buildout'].pop('extends', None)
+            if cl_extends:
+                for extends in cl_extends.value.split():
+                    download_options = for_download_options['buildout']
+                    cfg_data, _ = _open(
+                        os.path.dirname(extends),
+                        os.path.basename(extends),
+                        [], download_options,
+                        override, set(), user_defaults
+                    )
+                    data = _update(data, cfg_data)
 
         # apply command-line options
-        _update(data, cloptions)
+        data = _update(data, cloptions)
 
         # Set up versions section, if necessary
         if 'versions' not in data['buildout']:
@@ -440,7 +472,7 @@ class Buildout(DictMixin):
         # provide some defaults before options are parsed
         # because while parsing options those attributes might be
         # used already (Gottfried Ganssauge)
-        buildout_section = data['buildout']
+        buildout_section = self._raw['buildout']
 
         # Try to make sure we have absolute paths for standard
         # directories. We do this before doing substitutions, in case
@@ -606,9 +638,16 @@ class Buildout(DictMixin):
         ws = pkg_resources.WorkingSet(entries)
         ws.require('zc.buildout')
         options = self['buildout']
+        eggs_dir = options['eggs-directory']
+        develop_eggs_dir = options['develop-eggs-directory']
+        ws = zc.buildout.easy_install.sort_working_set(
+                ws,
+                eggs_dir=eggs_dir,
+                develop_eggs_dir=develop_eggs_dir
+                )
         zc.buildout.easy_install.scripts(
             ['zc.buildout'], ws, sys.executable,
-            self['buildout']['bin-directory'],
+            options['bin-directory'],
             relative_paths = (
                 bool_option(options, 'relative-paths', False)
                 and options['directory']
@@ -1090,11 +1129,14 @@ class Buildout(DictMixin):
         # If they do, do the upgrade and restart the buildout process.
         __doing__ = 'Checking for upgrades.'
 
+        if 'BUILDOUT_RESTART_AFTER_UPGRADE' in os.environ:
+            return
+
         if not self.newest:
             return
 
         ws = zc.buildout.easy_install.install(
-            ('zc.buildout', 'setuptools'),
+            ('zc.buildout', 'setuptools', 'pip', 'wheel'),
             self['buildout']['eggs-directory'],
             links = self['buildout'].get('find-links', '').split(),
             index = self['buildout'].get('index'),
@@ -1103,14 +1145,13 @@ class Buildout(DictMixin):
             )
 
         upgraded = []
-        # The setuptools/zc.buildout locations at the time we started was
-        # recorded in easy_install.py. We use that here to check if we've been
-        # upgraded.
-        start_locations = zc.buildout.easy_install.buildout_and_setuptools_path
-        for project in 'zc.buildout', 'setuptools':
+
+        for project in 'zc.buildout', 'setuptools', 'pip', 'wheel':
             req = pkg_resources.Requirement.parse(project)
-            if ws.find(req).location not in start_locations:
-                upgraded.append(ws.find(req))
+            dist = ws.find(req)
+            importlib.import_module(project)
+            if not inspect.getfile(sys.modules[project]).startswith(dist.location):
+                upgraded.append(dist)
 
         if not upgraded:
             return
@@ -1131,7 +1172,7 @@ class Buildout(DictMixin):
                                  "buildout command.")
             return
 
-        self._logger.info("Upgraded:\n  %s;\nrestarting.",
+        self._logger.info("Upgraded:\n  %s;\nRestarting.",
                           ",\n  ".join([("%s version %s"
                                        % (dist.project_name, dist.version)
                                        )
@@ -1143,9 +1184,16 @@ class Buildout(DictMixin):
         # the new dist is different, so we've upgraded.
         # Update the scripts and return True
         options = self['buildout']
+        eggs_dir = options['eggs-directory']
+        develop_eggs_dir = options['develop-eggs-directory']
+        ws = zc.buildout.easy_install.sort_working_set(
+                ws,
+                eggs_dir=eggs_dir,
+                develop_eggs_dir=develop_eggs_dir
+                )
         zc.buildout.easy_install.scripts(
             ['zc.buildout'], ws, sys.executable,
-            self['buildout']['bin-directory'],
+            options['bin-directory'],
             relative_paths = (
                 bool_option(options, 'relative-paths', False)
                 and options['directory']
@@ -1157,7 +1205,8 @@ class Buildout(DictMixin):
         if not __debug__:
             args.insert(0, '-O')
         args.insert(0, sys.executable)
-        sys.exit(subprocess.call(args))
+        env=dict(os.environ, BUILDOUT_RESTART_AFTER_UPGRADE='1')
+        sys.exit(subprocess.call(args, env=env))
 
     def _load_extensions(self):
         __doing__ = 'Loading extensions.'
@@ -1378,6 +1427,12 @@ def _install_and_load(spec, group, entry, buildout):
                 dest = buildout_options['eggs-directory']
                 path = [buildout_options['develop-eggs-directory']]
 
+            # Pin versions when processing the buildout section
+            versions_section_name = buildout['buildout'].get('versions', 'versions')
+            versions = buildout.get(versions_section_name, {})
+            zc.buildout.easy_install.allow_picked_versions(
+                bool_option(buildout['buildout'], 'allow-picked-versions')
+                )
             zc.buildout.easy_install.install(
                 [spec], dest,
                 links=buildout._links,
@@ -1385,7 +1440,8 @@ def _install_and_load(spec, group, entry, buildout):
                 path=path,
                 working_set=pkg_resources.working_set,
                 newest=buildout.newest,
-                allow_hosts=buildout._allow_hosts
+                allow_hosts=buildout._allow_hosts,
+                versions=versions,
                 )
 
         __doing__ = 'Loading %s recipe entry %s:%s.', group, spec, entry
@@ -1396,7 +1452,7 @@ def _install_and_load(spec, group, entry, buildout):
         v = sys.exc_info()[1]
         buildout._logger.log(
             1,
-            "Could't load %s entry point %s\nfrom %s:\n%s.",
+            "Couldn't load %s entry point %s\nfrom %s:\n%s.",
             group, entry, spec, v)
         raise
 
@@ -1464,7 +1520,7 @@ class Options(DictMixin):
 
             result = _annotate_section(result, "")
             data = _annotate_section(copy.deepcopy(data), "")
-            _update_section(result, data)
+            result = _update_section(result, data)
             result = _unannotate_section(result)
             result.pop('<', None)
             return result
@@ -1714,9 +1770,10 @@ def _default_globals():
     globals_defs.update({'python2': major_python_versions[0] == '2',
                          'python3': major_python_versions[0] == '3'})
 
-    # minor python major_python_versions as python24, python25 ... python36
+    # minor python major_python_versions as python24, python25 ... python39
     minor_python_versions = ('24', '25', '26', '27',
-                             '30', '31', '32', '33', '34', '35', '36')
+                             '30', '31', '32', '33', '34', '35', '36', '37', '38', '39',
+                             '310', '311', '312', '313', '314', '315')
     for v in minor_python_versions:
         globals_defs['python' + v] = ''.join(major_python_versions[:2]) == v
 
@@ -1751,18 +1808,28 @@ def _default_globals():
 
     return globals_defs
 
-def _open(base, filename, seen, dl_options, override, downloaded):
+variable_template_split = re.compile('([$]{[^}]*})').split
+
+def _open(
+        base, filename, seen, download_options,
+        override, downloaded, user_defaults
+        ):
     """Open a configuration file and return the result as a dictionary,
 
     Recursively open other files based on buildout options found.
     """
-    _update_section(dl_options, override)
-    dl_options_copy = copy.deepcopy(dl_options)
-    _dl_options = _unannotate_section(dl_options_copy)
-    newest = bool_option(_dl_options, 'newest', 'false')
+    download_options = _update_section(download_options, override)
+    raw_download_options = _unannotate_section(download_options)
+    newest = bool_option(raw_download_options, 'newest', 'false')
     fallback = newest and not (filename in downloaded)
+    extends_cache = raw_download_options.get('extends-cache')
+    if extends_cache and variable_template_split(extends_cache)[1::2]:
+        raise ValueError(
+            "extends-cache '%s' may not contain ${section:variable} to expand."
+            % extends_cache
+        )
     download = zc.buildout.download.Download(
-        _dl_options, cache=_dl_options.get('extends-cache'),
+        raw_download_options, cache=extends_cache,
         fallback=fallback, hash_name=True)
     is_temp = False
     downloaded_filename = None
@@ -1815,22 +1882,49 @@ def _open(base, filename, seen, dl_options, override, downloaded):
     result = _annotate(result, filename)
 
     if root_config_file and 'buildout' in result:
-        dl_options = _update_section(dl_options, result['buildout'])
+        download_options = _update_section(
+            download_options, result['buildout']
+        )
 
+    # Process extends to handle nested += and -=
+    eresults = []
     if extends:
         extends = extends.split()
-        eresult = _open(base, extends.pop(0), seen, dl_options, override,
-                        downloaded)
         for fname in extends:
-            _update(eresult, _open(base, fname, seen, dl_options, override,
-                    downloaded))
-        result = _update(eresult, result)
+            next_extend, user_defaults = _open(
+                base, fname, seen, download_options, override,
+                downloaded, user_defaults)
+            eresults.extend(next_extend)
+    else:
+        if user_defaults:
+            result = _update(user_defaults, result)
+            user_defaults = {}
 
+    optional_extends = options.pop('optional-extends', None)
+    if optional_extends:
+        optional_extends = optional_extends.value.split()
+        for fname in optional_extends:
+            if not os.path.exists(fname):
+                print("optional-extends file not found: %s" % fname)
+                continue
+            next_extend, user_defaults = _open(
+                base, fname, seen, download_options, override,
+                downloaded, user_defaults)
+            eresults.extend(next_extend)
+
+    eresults.append(result)
     seen.pop()
-    return result
+
+    if root_config_file:
+        final_result = {}
+        for eresult in eresults:
+            final_result = _update(final_result, eresult)
+        return final_result, user_defaults
+    else:
+        return eresults, user_defaults
 
 
-ignore_directories = '.svn', 'CVS', '__pycache__'
+ignore_directories = '.svn', 'CVS', '__pycache__', '.git'
 _dir_hashes = {}
 def _dir_hash(dir):
     dir_hash = _dir_hashes.get(dir, None)
@@ -1882,19 +1976,21 @@ def _dists_sig(dists):
             result.append(os.path.basename(location))
     return result
 
-def _update_section(s1, s2):
+def _update_section(in1, s2):
+    s1 = copy.deepcopy(in1)
     # Base section 2 on section 1; section 1 is copied, with key-value pairs
     # in section 2 overriding those in section 1. If there are += or -=
-    # operators in section 2, process these to add or substract items (delimited
+    # operators in section 2, process these to add or subtract items (delimited
     # by newlines) from the preexisting values.
     s2 = copy.deepcopy(s2) # avoid mutating the second argument, which is unexpected
-    # Sort on key, then on the addition or substraction operator (+ comes first)
+    # Sort on key, then on the addition or subtraction operator (+ comes first)
     for k, v in sorted(s2.items(), key=lambda x: (x[0].rstrip(' +'), x[0][-1])):
         if k.endswith('+'):
             key = k.rstrip(' +')
             implicit_value = SectionKey("", "IMPLICIT_VALUE")
             # Find v1 in s2 first; it may have been defined locally too.
             section_key = s2.get(key, s1.get(key, implicit_value))
+            section_key = copy.deepcopy(section_key)
             section_key.addToValue(v.value, v.source)
             s2[key] = section_key
             del s2[k]
@@ -1903,6 +1999,7 @@ def _update_section(s1, s2):
             implicit_value = SectionKey("", "IMPLICIT_VALUE")
             # Find v1 in s2 first; it may have been set by a += operation first
             section_key = s2.get(key, s1.get(key, implicit_value))
+            section_key = copy.deepcopy(section_key)
             section_key.removeFromValue(v.value, v.source)
             s2[key] = section_key
             del s2[k]
@@ -1916,14 +2013,39 @@ def _update_verbose(s1, s2):
             v1 = s1[key]
             v1.overrideValue(v2)
         else:
-            s1[key] = v2
+            s1[key] = copy.deepcopy(v2)
 
-def _update(d1, d2):
+def _update(in1, d2):
+    d1 = copy.deepcopy(in1)
     for section in d2:
         if section in d1:
             d1[section] = _update_section(d1[section], d2[section])
+        elif '<' not in d2[section].keys():
+            # Skip sections that extend in other sections (macros), as we don't
+            # have all the data (these will be processed when the section is
+            # extended)
+            temp = copy.deepcopy(d2[section])
+            # 641 - Process base definitions done with += and -=
+            for k, v in sorted(temp.items(), key=lambda item: item[0]):
+                # Process + before -, configparser resolves conflicts
+                if k[-1] == '+' and k[:-2] not in temp:
+                    # Turn += without a preceding = into an assignment
+                    temp[k[:-2]] = temp[k]
+                    del temp[k]
+                elif k[-1] == '-' and k[:-2] not in temp:
+                    # Turn -= without a preceding = into an empty assignment
+                    temp[k[:-2]] = temp[k]
+                    temp[k[:-2]].removeFromValue(
+                        temp[k[:-2]].value, "IMPLICIT_VALUE"
+                        )
+                    del temp[k]
+
+            # 656 - Handle multiple option assignments/extensions/removals
+            # in the same file, which can happen with conditional sections
+            d1[section] = _update_section({}, temp)
         else:
-            d1[section] = d2[section]
+            d1[section] = copy.deepcopy(d2[section])
+
     return d1
 
 def _recipe(options):
@@ -1966,8 +2088,11 @@ def _check_for_unused_options_in_section(buildout, section):
     unused = [option for option in sorted(options._raw)
               if option not in options._data]
     if unused:
-        buildout._logger.warning("Unused options for %s: %s."
-                                 % (section, ' '.join(map(repr, unused)))
+        buildout._logger.warning(
+            "Section `%s` contains unused option(s): %s.\n"
+            "This may be an indication for either a typo in the option's name "
+            "or a bug in the used recipe." %
+            (section, ' '.join(map(repr, unused)))
         )
 
 _usage = """\
@@ -1984,7 +2109,7 @@ Options:
   -D
 
     Debug errors.  If an error occurs, then the post-mortem debugger
-    will be started. This is especially useful for debuging recipe
+    will be started. This is especially useful for debugging recipe
     problems.
 
   -h, --help
@@ -2067,7 +2192,7 @@ Commands:
     COMPUTED_VALUE, DEFAULT_VALUE, COMMAND_LINE_VALUE).
 
   query section:key
-    
+
     Display value of given section key pair.
 """
 
@@ -2088,7 +2213,7 @@ def main(args=None):
     config_file = 'buildout.cfg'
     verbosity = 0
     options = []
-    user_defaults = True
+    use_user_defaults = True
     debug = False
     while args:
         if args[0][0] == '-':
@@ -2100,7 +2225,7 @@ def main(args=None):
                 elif op[0] == 'q':
                     verbosity -= 10
                 elif op[0] == 'U':
-                    user_defaults = False
+                    use_user_defaults = False
                 elif op[0] == 'o':
                     options.append(('buildout', 'offline', 'true'))
                 elif op[0] == 'O':
@@ -2142,7 +2267,8 @@ def main(args=None):
                     _help()
                 elif orig_op == '--version':
                     _version()
-                _error("Invalid option", '-'+op[0])
+                else:
+                    _error("Invalid option", '-'+op[0])
         elif '=' in args[0]:
             option, value = args.pop(0).split('=', 1)
             option = option.split(':')
@@ -2170,7 +2296,7 @@ def main(args=None):
     try:
         try:
             buildout = Buildout(config_file, options,
-                                user_defaults, command, args)
+                                use_user_defaults, command, args)
             getattr(buildout, command)(args)
         except SystemExit:
             logging.shutdown()
