@@ -18,10 +18,12 @@ It doesn't install scripts.  It uses setuptools and requires it to be
 installed.
 """
 
+import copy
 import distutils.errors
 import errno
 import glob
 import logging
+import operator
 import os
 import pkg_resources
 import py_compile
@@ -36,7 +38,13 @@ import sys
 import tempfile
 import zc.buildout
 import zc.buildout.rmtree
+from zc.buildout import WINDOWS
+from zc.buildout import PY3
+from zc.buildout._compat import packaging_utils
+from zc.buildout._compat import specifiers
+from zc.buildout.utils import normalize_name
 import warnings
+import csv
 
 try:
     from setuptools.wheel import Wheel  # This is the important import
@@ -48,6 +56,9 @@ try:
 except ImportError:
     SETUPTOOLS_SUPPORTS_WHEELS = False
 
+
+BIN_SCRIPTS = 'Scripts' if WINDOWS else 'bin'
+
 warnings.filterwarnings(
     'ignore', '.+is being parsed as a legacy, non PEP 440, version')
 
@@ -56,7 +67,7 @@ def realpath(path):
     return os.path.normcase(os.path.abspath(_oprp(path)))
 
 default_index_url = os.environ.get(
-    'buildout-testing-index-url',
+    'buildout_testing_index_url',
     'https://pypi.org/simple',
     )
 
@@ -80,19 +91,24 @@ has_distribute = pkg_resources.working_set.find(
 has_setuptools = pkg_resources.working_set.find(
         pkg_resources.Requirement.parse('setuptools')) is not None
 if has_distribute and not has_setuptools:
-    sys.exit("zc.buildout 2 needs setuptools, not distribute."
-             "  Are you using an outdated bootstrap.py?  Make sure"
-             " you have the latest version downloaded from"
-             " https://bootstrap.pypa.io/bootstrap-buildout.py")
+    sys.exit("zc.buildout 3 needs setuptools, not distribute."
+             "Did you properly install with pip in a virtualenv ?")
 
 # Include buildout and setuptools eggs in paths.  We get this
 # initially from the entire working set.  Later, we'll use the install
 # function to narrow to just the buildout and setuptools paths.
-buildout_and_setuptools_path = [d.location for d in pkg_resources.working_set]
+buildout_and_setuptools_path = sorted({d.location for d in pkg_resources.working_set})
 setuptools_path = buildout_and_setuptools_path
+pip_path = buildout_and_setuptools_path
+logger.debug('before restricting versions: pip_path %r', pip_path)
 
 FILE_SCHEME = re.compile('file://', re.I).match
 DUNDER_FILE_PATTERN = re.compile(r"__file__ = '(?P<filename>.+)'$")
+BUILDOUT_PIP_NO_BUILD_ISOLATION = os.getenv("BUILDOUT_PIP_NO_BUILD_ISOLATION")
+if BUILDOUT_PIP_NO_BUILD_ISOLATION and BUILDOUT_PIP_NO_BUILD_ISOLATION.lower() in (
+    "0", "false", "no", "off"
+):
+    BUILDOUT_PIP_NO_BUILD_ISOLATION=False
 
 class _Monkey(object):
     def __init__(self, module, **kw):
@@ -114,7 +130,55 @@ class _NoWarn(object):
 
 _no_warn = _NoWarn()
 
-class AllowHostsPackageIndex(setuptools.package_index.PackageIndex):
+
+class EnvironmentMixin(object):
+    """Mixin class for Environment and PackageIndex for canonicalized names.
+
+    * pkg_resources defines the Environment class
+    * setuptools defines a PackageIndex class that inherits from Environment
+    * Buildout needs a few fixes that should be used by both.
+
+    The fixes are needed for this issue, where distributions created by
+    setuptools 69.3+ get a different name than with older versions:
+    https://github.com/buildout/buildout/issues/647
+    """
+    def __getitem__(self, project_name):
+        """Return a newest-to-oldest list of distributions for `project_name`
+
+        Uses case-insensitive `project_name` comparison, assuming all the
+        project's distributions use their project's name converted to all
+        lowercase as their key.
+
+        """
+        distribution_key = normalize_name(project_name)
+        return self._distmap.get(distribution_key, [])
+
+    def add(self, dist):
+        """Add `dist` if we ``can_add()`` it and it has not already been added
+        """
+        if self.can_add(dist) and dist.has_version():
+            # Instead of 'dist.key' we add a normalized version.
+            distribution_key = normalize_name(dist.key)
+            dists = self._distmap.setdefault(distribution_key, [])
+            if dist not in dists:
+                dists.append(dist)
+                dists.sort(key=operator.attrgetter('hashcmp'), reverse=True)
+
+
+class Environment(EnvironmentMixin, pkg_resources.Environment):
+    """Buildout version of Environment with canonicalized names.
+
+    * pkg_resources defines the Environment class
+    * setuptools defines a PackageIndex class that inherits from Environment
+    * Buildout needs a few fixes that should be used by both.
+
+    The fixes are needed for this issue, where distributions created by
+    setuptools 69.3+ get a different name than with older versions:
+    https://github.com/buildout/buildout/issues/647
+    """
+
+
+class AllowHostsPackageIndex(EnvironmentMixin, setuptools.package_index.PackageIndex):
     """Will allow urls that are local to the system.
 
     No matter what is allow_hosts.
@@ -175,7 +239,6 @@ def _execute_permission():
     return 0o777 - current_umask
 
 
-_easy_install_cmd = 'from setuptools.command.easy_install import main; main()'
 
 def get_namespace_package_paths(dist):
     """
@@ -283,7 +346,7 @@ class Installer(object):
 
     def _make_env(self):
         full_path = self._get_dest_dist_paths() + self._path
-        env = pkg_resources.Environment(full_path)
+        env = Environment(full_path)
         # this needs to be called whenever self._env is modified (or we could
         # make an Environment subclass):
         self._eggify_env_dest_dists(env, self._dest)
@@ -317,7 +380,7 @@ class Installer(object):
         """Return textual requirements/constraint information for debug purposes
 
         We do a very simple textual search, as that filters out most
-        extraneous information witout missing anything.
+        extraneous information without missing anything.
 
         """
         output = [
@@ -382,7 +445,7 @@ class Installer(object):
             logger.debug(
                 'There are no distros available that meet %r.\n'
                 'Using our best, %s.',
-                str(req), best_available)
+                str(req), best_we_have)
             return best_we_have, None
 
         if self._prefer_final:
@@ -416,14 +479,14 @@ class Installer(object):
             str(req))
         return best_we_have, None
 
-    def _call_easy_install(self, spec, dest, dist):
+    def _call_pip_install(self, spec, dest, dist):
 
         tmp = tempfile.mkdtemp(dir=dest)
         try:
-            paths = call_easy_install(spec, tmp)
+            paths = call_pip_install(spec, tmp)
 
             dists = []
-            env = pkg_resources.Environment(paths)
+            env = Environment(paths)
             for project in env:
                 dists.extend(env[project])
 
@@ -539,7 +602,10 @@ class Installer(object):
         __doing__ = 'Getting distribution for %r.', str(requirement)
 
         # Maybe an existing dist is already the best dist that satisfies the
-        # requirement
+        # requirement.  If not, get a link to an available distribution that
+        # we could download.  The method returns a tuple with an existing
+        # dist or an available dist.  Either 'dist' is None, or 'avail'
+        # is None, or both are None.
         dist, avail = self._satisfied(requirement)
 
         if dist is None:
@@ -551,9 +617,8 @@ class Installer(object):
 
             logger.info(*__doing__)
 
-            # Retrieve the dist:
             if avail is None:
-                self._index.obtain(requirement)
+                # We have no existing dist, and none is available for download.
                 raise MissingDistribution(requirement, ws)
 
             # We may overwrite distributions, so clear importer
@@ -628,10 +693,11 @@ class Installer(object):
                 self._picked_versions[dist.project_name] = dist.version
 
                 if not self._allow_picked_versions:
-                    raise zc.buildout.UserError(
-                        'Picked: %s = %s' % (dist.project_name,
-                                             dist.version)
-                        )
+                    msg = NOT_PICKED_AND_NOT_ALLOWED.format(
+                        name=dist.project_name,
+                        version=dist.version
+                    )
+                    raise zc.buildout.UserError(msg)
 
     def _maybe_add_setuptools(self, ws, dist):
         if dist_needs_pkg_resources(dist):
@@ -670,9 +736,15 @@ class Installer(object):
 
         for_buildout_run = bool(working_set)
 
-        requirements = [self._constrain(pkg_resources.Requirement.parse(spec))
+        requirements = [pkg_resources.Requirement.parse(spec)
                         for spec in specs]
 
+        requirements = [
+            self._constrain(requirement)
+            for requirement in requirements
+            if not requirement.marker or requirement.marker.evaluate()
+        ]
+
         if working_set is None:
             ws = pkg_resources.WorkingSet([])
         else:
@@ -695,7 +767,7 @@ class Installer(object):
         # Note that we don't use the existing environment, because we want
         # to look for new eggs unless what we have is the best that
         # matches the requirement.
-        env = pkg_resources.Environment(ws.entries)
+        env = Environment(ws.entries)
 
         while requirements:
             # Process dependencies breadth-first.
@@ -831,7 +903,7 @@ class Installer(object):
                 setuptools.command.setopt.edit_config(
                     setup_cfg, dict(build_ext=build_ext))
 
-                dists = self._call_easy_install(base, self._dest, dist)
+                dists = self._call_pip_install(base, self._dest, dist)
 
                 return [dist.location for dist in dists]
             finally:
@@ -958,11 +1030,17 @@ def install(specs, dest,
 
 buildout_and_setuptools_dists = list(install(['zc.buildout'], None,
                                              check_picked=False))
-buildout_and_setuptools_path = [d.location
-                                for d in buildout_and_setuptools_dists]
-setuptools_path = [d.location
-                   for d in install(['setuptools'], None, check_picked=False)]
-setuptools_pythonpath = os.pathsep.join(setuptools_path)
+buildout_and_setuptools_path = sorted({d.location
+                                for d in buildout_and_setuptools_dists})
+
+pip_dists = [d for d in buildout_and_setuptools_dists if d.project_name != 'zc.buildout']
+pip_path = sorted({d.location for d in pip_dists})
+logger.debug('after restricting versions: pip_path %r', pip_path)
+pip_pythonpath = os.pathsep.join(pip_path)
+
+setuptools_path = pip_path
+setuptools_pythonpath = pip_pythonpath
+
 
 def build(spec, dest, build_ext,
           links=(), index=None,
@@ -1138,7 +1216,7 @@ def scripts(reqs, working_set, executable, dest=None,
     for p in path:
         if p not in unique_path:
             unique_path.append(p)
-    path = list(map(realpath, unique_path))
+    path = [realpath(p) for p in unique_path]
 
     generated = []
 
@@ -1154,7 +1232,10 @@ def scripts(reqs, working_set, executable, dest=None,
     for req in reqs:
         if isinstance(req, str):
             req = pkg_resources.Requirement.parse(req)
+            if req.marker and not req.marker.evaluate():
+                continue
             dist = working_set.find(req)
+            assert dist is not None, 'distribution not found for %s' % req.name
             # regular console_scripts entry points
             for name in pkg_resources.get_entry_map(dist, 'console_scripts'):
                 entry_point = dist.get_entry_info('console_scripts', name)
@@ -1188,7 +1269,10 @@ def scripts(reqs, working_set, executable, dest=None,
         else:
             entry_points.append(req)
 
+    entry_points_names = []
+
     for name, module_name, attrs in entry_points:
+        entry_points_names.append(name)
         if scripts is not None:
             sname = scripts.get(name)
             if sname is None:
@@ -1204,6 +1288,18 @@ def scripts(reqs, working_set, executable, dest=None,
                     initialization, rpsetup)
             )
 
+    # warn when a script name passed in 'scripts' argument
+    # is not defined in an entry point.
+    if scripts is not None:
+        for name, target in scripts.items():
+            if name not in entry_points_names:
+                if name == target:
+                    logger.warning("Could not generate script '%s' as it is not "
+                        "defined in the egg entry points.", name)
+                else:
+                    logger.warning("Could not generate script '%s' as script "
+                        "'%s' is not defined in the egg entry points.", name, target)
+
     for name, contents in distutils_scripts:
         if scripts is not None:
             sname = scripts.get(name)
@@ -1484,7 +1580,9 @@ sys.path[0:0] = [
 
 _interactive = True
 if len(sys.argv) > 1:
-    _options, _args = __import__("getopt").getopt(sys.argv[1:], 'ic:m:')
+    # The Python interpreter wrapper allows only some of the options that a
+    # "regular" Python interpreter accepts.
+    _options, _args = __import__("getopt").getopt(sys.argv[1:], 'Iic:m:')
     _interactive = False
     for (_opt, _val) in _options:
         if _opt == '-i':
@@ -1496,6 +1594,15 @@ if len(sys.argv) > 1:
             _args = []
             __import__("runpy").run_module(
                  _val, {}, "__main__", alter_sys=True)
+        elif _opt == '-I':
+            # Allow yet silently ignore the `-I` option. The original behaviour
+            # for this option is to create an isolated Python runtime. It was
+            # deemed acceptable to allow the option here as this Python wrapper
+            # is isolated from the system Python already anyway.
+            # The specific use-case that led to this change is how the Python
+            # language extension for Visual Studio Code calls the Python
+            # interpreter when initializing the extension.
+            pass
 
     if _args:
         sys.argv[:] = _args
@@ -1593,30 +1700,25 @@ def redo_pyc(egg):
                 call_subprocess(args)
 
 def _constrained_requirement(constraint, requirement):
+    assert isinstance(requirement, pkg_resources.Requirement)
     if constraint[0] not in '<>':
         if constraint.startswith('='):
             assert constraint.startswith('==')
-            constraint = constraint[2:]
-        if constraint not in requirement:
+            version = constraint[2:]
+        else:
+            version = constraint
+            constraint = '==' + constraint
+        if version not in requirement:
             msg = ("The requirement (%r) is not allowed by your [versions] "
-                   "constraint (%s)" % (str(requirement), constraint))
+                   "constraint (%s)" % (str(requirement), version))
             raise IncompatibleConstraintError(msg)
-
-        # Sigh, copied from Requirement.__str__
-        extras = ','.join(requirement.extras)
-        if extras:
-            extras = '[%s]' % extras
-        return pkg_resources.Requirement.parse(
-            "%s%s==%s" % (requirement.project_name, extras, constraint))
-
-    if requirement.specs:
-        return pkg_resources.Requirement.parse(
-            str(requirement) + ',' + constraint
-            )
+        specifier = specifiers.SpecifierSet(constraint)
     else:
-        return pkg_resources.Requirement.parse(
-            str(requirement) + ' ' + constraint
-            )
+        specifier = requirement.specifier & constraint
+    constrained = copy.deepcopy(requirement)
+    constrained.specifier = specifier
+    return pkg_resources.Requirement.parse(str(constrained))
+
 
 class IncompatibleConstraintError(zc.buildout.UserError):
     """A specified version is incompatible with a given requirement.
@@ -1625,40 +1727,177 @@ class IncompatibleConstraintError(zc.buildout.UserError):
 IncompatibleVersionError = IncompatibleConstraintError # Backward compatibility
 
 
-def call_easy_install(spec, dest):
+def call_pip_install(spec, dest):
     """
-    Call `easy_install` from setuptools as a subprocess to install a
+    Call `pip install` from a subprocess to install a
     distribution specified by `spec` into `dest`.
     Returns all the paths inside `dest` created by the above.
     """
-    path = setuptools_path
-
-    args = [sys.executable, '-c',
-            ('import sys; sys.path[0:0] = %r; ' % path) +
-            _easy_install_cmd, '-mZUNxd', dest]
+    args = [sys.executable, '-m', 'pip', 'install', '--no-deps', '-t', dest]
+    if BUILDOUT_PIP_NO_BUILD_ISOLATION:
+        args.append('--no-build-isolation')
     level = logger.getEffectiveLevel()
-    if level > 0:
+    if level >= logging.INFO:
         args.append('-q')
-    elif level < 0:
+    else:
         args.append('-v')
 
     args.append(spec)
 
+    try:
+        from pip._internal.cli.cmdoptions import no_python_version_warning
+        HAS_WARNING_OPTION = True
+    except ImportError:
+        HAS_WARNING_OPTION = False
+    if HAS_WARNING_OPTION:
+        if not hasattr(call_pip_install, 'displayed'):
+            call_pip_install.displayed = True
+        else:
+            args.append('--no-python-version-warning')
+
+    env = copy.copy(os.environ)
+    python_path = copy.copy(pip_path)
+    python_path.append(env.get('PYTHONPATH', ''))
+    env['PYTHONPATH'] = os.pathsep.join(python_path)
+
     if level <= logging.DEBUG:
-        logger.debug('Running easy_install:\n"%s"\npath=%s\n',
-                        '" "'.join(args), path)
+        logger.debug('Running pip install:\n"%s"\npath=%s\n',
+                        '" "'.join(args), pip_path)
 
     sys.stdout.flush() # We want any pending output first
 
-    exit_code = subprocess.call(list(args))
+    exit_code = subprocess.call(list(args), env=env)
 
     if exit_code:
         logger.error(
             "An error occurred when trying to install %s. "
             "Look above this message for any errors that "
-            "were output by easy_install.",
+            "were output by pip install.",
             spec)
-    return glob.glob(os.path.join(dest, '*'))
+        sys.exit(1)
+
+    split_entries = [os.path.splitext(entry) for entry in os.listdir(dest)]
+    try:
+        distinfo_dir = [
+            base + ext for base, ext in split_entries if ext == ".dist-info"
+        ][0]
+    except IndexError:
+        logger.error(
+            "No .dist-info directory after successful pip install of %s",
+            spec)
+        raise
+
+    return make_egg_after_pip_install(dest, distinfo_dir)
+
+
+def make_egg_after_pip_install(dest, distinfo_dir):
+    """build properly named egg directory"""
+
+    # `pip install` does not build the namespace aware __init__.py files
+    # but they are needed in egg directories.
+    # Add them before moving files setup by pip
+    namespace_packages_file = os.path.join(
+        dest, distinfo_dir,
+        'namespace_packages.txt'
+    )
+    if os.path.isfile(namespace_packages_file):
+        with open(namespace_packages_file) as f:
+            namespace_packages = [
+                line.strip().replace('.', os.path.sep)
+                for line in f.readlines()
+            ]
+
+        for namespace_package in namespace_packages:
+            namespace_package_dir = os.path.join(dest, namespace_package)
+            if os.path.isdir(namespace_package_dir):
+                init_py_file = os.path.join(
+                    namespace_package_dir, '__init__.py')
+                with open(init_py_file, 'w') as f:
+                    f.write(
+                        "__import__('pkg_resources')."
+                        "declare_namespace(__name__)"
+                    )
+
+    # Remove `bin` directory if needed
+    # as there is no way to avoid script installation
+    # when running `pip install`
+    entry_points_file = os.path.join(dest, distinfo_dir, 'entry_points.txt')
+    if os.path.isfile(entry_points_file):
+        with open(entry_points_file) as f:
+            content = f.read()
+            if "console_scripts" in content or "gui_scripts" in content:
+                bin_dir = os.path.join(dest, BIN_SCRIPTS)
+                if os.path.exists(bin_dir):
+                    shutil.rmtree(bin_dir)
+
+    # Make properly named new egg dir
+    distro = list(pkg_resources.find_distributions(dest))[0]
+    base = "{}-{}".format(
+        distro.egg_name(), pkg_resources.get_supported_platform()
+    )
+    egg_name = base + '.egg'
+    new_distinfo_dir = base + '.dist-info'
+    egg_dir = os.path.join(dest, egg_name)
+    os.mkdir(egg_dir)
+
+    # Move ".dist-info" dir into new egg dir
+    os.rename(
+        os.path.join(dest, distinfo_dir),
+        os.path.join(egg_dir, new_distinfo_dir)
+    )
+
+    top_level_file = os.path.join(egg_dir, new_distinfo_dir, 'top_level.txt')
+    if os.path.isfile(top_level_file):
+        with open(top_level_file) as f:
+            top_levels = filter(
+                (lambda x: len(x) != 0),
+                [line.strip() for line in f.readlines()]
+                )
+    else:
+        top_levels = ()
+
+    # Move all top_level modules or packages
+    for top_level in top_levels:
+        # as package
+        top_level_dir = os.path.join(dest, top_level)
+        if os.path.exists(top_level_dir):
+            shutil.move(top_level_dir, egg_dir)
+            continue
+        # as module
+        top_level_py = top_level_dir + '.py'
+        if os.path.exists(top_level_py):
+            shutil.move(top_level_py, egg_dir)
+            top_level_pyc = top_level_dir + '.pyc'
+            if os.path.exists(top_level_pyc):
+                shutil.move(top_level_pyc, egg_dir)
+            continue
+
+    record_file = os.path.join(egg_dir, new_distinfo_dir, 'RECORD')
+    if os.path.isfile(record_file):
+        if PY3:
+            with open(record_file, newline='') as f:
+                all_files = [row[0] for row in csv.reader(f)]
+        else:
+            with open(record_file, 'rb') as f:
+                all_files = [row[0] for row in csv.reader(f)]
+
+    # There might be some c extensions left over
+    for entry in all_files:
+        if entry.endswith(('.pyc', '.pyo')):
+            continue
+        dest_entry = os.path.join(dest, entry)
+        # work around pip install -t bug that leaves entries in RECORD
+        # that starts with '../../'
+        if not os.path.abspath(dest_entry).startswith(dest):
+            continue
+        egg_entry = os.path.join(egg_dir, entry)
+        if os.path.exists(dest_entry) and not os.path.exists(egg_entry):
+            egg_entry_dir = os.path.dirname(egg_entry)
+            if not os.path.exists(egg_entry_dir):
+                os.makedirs(egg_entry_dir)
+            os.rename(dest_entry, egg_entry)
+
+    return [egg_dir]
 
 
 def unpack_egg(location, dest):
@@ -1699,10 +1938,10 @@ def _get_matching_dist_in_location(dist, location):
     # may be normalized (e.g., 3.3 becomes 3.3.0 when downloaded from
     # PyPI.)
 
-    env = pkg_resources.Environment([location])
+    env = Environment([location])
     dists = [ d for project_name in env for d in env[project_name] ]
-    dist_infos = [ (d.project_name.lower(), d.parsed_version) for d in dists ]
-    if dist_infos == [(dist.project_name.lower(), dist.parsed_version)]:
+    dist_infos = [ (normalize_name(d.project_name), d.parsed_version) for d in dists ]
+    if dist_infos == [(normalize_name(dist.project_name), dist.parsed_version)]:
         return dists.pop()
 
 def _move_to_eggs_dir_and_compile(dist, dest):
@@ -1730,6 +1969,7 @@ def _move_to_eggs_dir_and_compile(dist, dest):
             raise
     tmp_dest = tempfile.mkdtemp(dir=dest)
     try:
+        installed_with_pip = False
         if (os.path.isdir(dist.location) and
                 dist.precedence >= pkg_resources.BINARY_DIST):
             # We got a pre-built directory. It must have been obtained locally.
@@ -1740,9 +1980,13 @@ def _move_to_eggs_dir_and_compile(dist, dest):
             # It is an archive of some sort.
             # Figure out how to unpack it, or fall back to easy_install.
             _, ext = os.path.splitext(dist.location)
-            unpacker = UNPACKERS.get(ext, call_easy_install)
-            unpacker(dist.location, tmp_dest)
-            [tmp_loc] = glob.glob(os.path.join(tmp_dest, '*'))
+            if ext in UNPACKERS:
+                unpacker = UNPACKERS[ext]
+                unpacker(dist.location, tmp_dest)
+                [tmp_loc] = glob.glob(os.path.join(tmp_dest, '*'))
+            else:
+                [tmp_loc] = call_pip_install(dist.location, tmp_dest)
+                installed_with_pip = True
 
         # We have installed the dist. Now try to rename/move it.
         newloc = os.path.join(dest, os.path.basename(tmp_loc))
@@ -1777,4 +2021,45 @@ def _move_to_eggs_dir_and_compile(dist, dest):
     finally:
         # Remember that temporary directories must be removed
         zc.buildout.rmtree.rmtree(tmp_dest)
+    if installed_with_pip:
+        newdist.precedence = pkg_resources.EGG_DIST
     return newdist
+
+
+def sort_working_set(ws, eggs_dir, develop_eggs_dir):
+    develop_paths = set()
+    pattern = os.path.join(develop_eggs_dir, '*.egg-link')
+    for egg_link in glob.glob(pattern):
+        with open(egg_link, 'rt') as f:
+            path = f.readline().strip()
+            if path:
+                develop_paths.add(path)
+
+    sorted_paths = []
+    egg_paths = []
+    other_paths = []
+    for dist in ws:
+        path = dist.location
+        if path in develop_paths:
+            sorted_paths.append(path)
+        elif os.path.commonprefix([path, eggs_dir]) == eggs_dir:
+            egg_paths.append(path)
+        else:
+            other_paths.append(path)
+    sorted_paths.extend(egg_paths)
+    sorted_paths.extend(other_paths)
+    return pkg_resources.WorkingSet(sorted_paths)
+
+
+NOT_PICKED_AND_NOT_ALLOWED = """\
+Picked: {name} = {version}
+
+The `{name}` egg does not have a version pin and `allow-picked-versions = false`.
+
+To resolve this, add
+
+    {name} = {version}
+
+to the [versions] section,
+
+OR set `allow-picked-versions = true`."""
