@@ -25,7 +25,7 @@ except ImportError:
 
 import errno
 import logging
-from multiprocessing import Process
+import multiprocessing
 import os
 import pkg_resources
 import random
@@ -125,12 +125,14 @@ def clean_up_pyc(*path):
 ## FIXME - check for other platforms
 MUST_CLOSE_FDS = not sys.platform.startswith('win')
 
-def system(command, input='', with_exit_code=False):
-    # Some TERMinals, expecially xterm and its variants, add invisible control
+def system(command, input='', with_exit_code=False, env=None):
+    # Some TERMinals, especially xterm and its variants, add invisible control
     # characters, which we do not want as they mess up doctests.  See:
     # https://github.com/buildout/buildout/pull/311
     # http://bugs.python.org/issue19884
-    env = dict(os.environ, TERM='dumb')
+    sub_env = dict(os.environ, TERM='dumb')
+    if env is not None:
+        sub_env.update(env)
 
     p = subprocess.Popen(command,
                          shell=True,
@@ -138,7 +140,7 @@ def system(command, input='', with_exit_code=False):
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          close_fds=MUST_CLOSE_FDS,
-                         env=env)
+                         env=sub_env)
     i, o, e = (p.stdin, p.stdout, p.stderr)
     if input:
         i.write(input.encode())
@@ -168,7 +170,7 @@ def _runsetup(setup, *args):
         zc.buildout.easy_install.call_subprocess(
             [sys.executable, setup] + args,
             env=dict(os.environ,
-                     PYTHONPATH=zc.buildout.easy_install.setuptools_pythonpath,
+                     PYTHONPATH=zc.buildout.easy_install.pip_pythonpath,
                      ),
             )
         if os.path.exists('build'):
@@ -265,7 +267,7 @@ def buildoutSetUp(test):
     register_teardown(lambda: rmtree(tmp))
 
     zc.buildout.easy_install.default_index_url = 'file://'+tmp
-    os.environ['buildout-testing-index-url'] = (
+    os.environ['buildout_testing_index_url'] = (
         zc.buildout.easy_install.default_index_url)
 
     def tmpdir(name):
@@ -297,15 +299,8 @@ def buildoutSetUp(test):
     # way due to the trick above:
     os.mkdir('develop-eggs')
 
-    if os.getenv("COVERAGE_PROCESS_START"):
-        # The user has requested subprocess code coverage. Since we will be changing
-        # directories, we need to make sure this path is absolute, which means
-        # we need to temporarily return to our starting directory.
-        os.chdir(here)
-        path_to_coveragerc = os.path.abspath(os.environ['COVERAGE_PROCESS_START'])
-        os.chdir(sample)
-        assert os.path.isfile(path_to_coveragerc), path_to_coveragerc
-        os.environ['COVERAGE_PROCESS_START'] = path_to_coveragerc
+    path_to_coveragerc = os.getenv("COVERAGE_PROCESS_START", None)
+    if path_to_coveragerc is not None:
 
         # Before we return to the current directory and destroy the
         # temporary working directory, we need to copy all the coverage files
@@ -586,7 +581,7 @@ else:
         re.compile(r'(\n?)-  \S+\.pyc\n'), '\\1')
 
 normalize_egg_py = (
-    re.compile(r'-py\d[.]\d(-\S+)?.egg'),
+    re.compile(r'-py\d[.]\d+(-\S+)?\.egg'),
     '-pyN.N.egg',
     )
 
@@ -599,6 +594,25 @@ normalize_open_in_generated_script = (
 
 not_found = (re.compile(r'Not found: [^\n]+/(\w|\.)+/\r?\n'), '')
 
+python27_warning = (re.compile(r'DEPRECATION: Python 2.7 reached the end of its '
+    'life on January 1st, 2020. Please upgrade your Python as Python 2.7 is no '
+    'longer maintained. A future version of pip will drop support for Python '
+    '2.7. More details about Python 2 support in pip, can be found at '
+    'https://pip.pypa.io/en/latest/development/release-process/#python-2-support\n'),
+    '')
+
+python27_warning_2 = (re.compile(r'DEPRECATION: Python 2.7 reached the end of its '
+    'life on January 1st, 2020. Please upgrade your Python as Python 2.7 is no '
+    'longer maintained. pip 21.0 will drop support for Python 2.7 in January 2021. '
+    'More details about Python 2 support in pip, can be found at '
+    'https://pip.pypa.io/en/latest/development/release-process/#python-2-support\n'),
+    '')
+
+easyinstall_deprecated = (re.compile(r'.*EasyInstallDeprecationWarning.*\n'),'')
+setuptools_deprecated = (re.compile(r'.*SetuptoolsDeprecationWarning.*\n'),'')
+pkg_resources_deprecated = (re.compile(r'.*PkgResourcesDeprecationWarning.*\n'),'')
+warnings_warn = (re.compile(r'.*warnings\.warn.*\n'),'')
+
 # Setuptools now pulls in dependencies when installed.
 adding_find_link = (re.compile(r"Adding find link '[^']+'"
                                r" from setuptools .*\r?\n"), '')
@@ -608,11 +622,35 @@ ignore_not_upgrading = (
     'Not upgrading because not running a local buildout command.\n'
     ), '')
 
+# The root logger from setuptools prints all kinds of lines.
+# This might depend on which setuptools version, or something else,
+# because it did not happen before.  Sample lines:
+# "root: Couldn't retrieve index page for 'zc.recipe.egg'"
+# "root: Scanning index of all packages.
+# "root: Found: /sample-buildout/recipe/dist/spam-2-pyN.N.egg"
+# I keep finding new lines like that, so let's ignore all.
+ignore_root_logger = (re.compile(r'root:.*'), '')
+# Now replace a multiline warning about that you should switch to native namespaces.
+ignore_native_namespace_warning_1 = (re.compile(r'!!'), '')
+ignore_native_namespace_warning_2 = (re.compile(r'\*' * 80), '')
+ignore_native_namespace_warning_3 = (re.compile(
+    r'Please replace its usage with implicit namespaces \(PEP 420\).'),
+    ''
+)
+ignore_native_namespace_warning_4 = (re.compile(
+    r'See https://setuptools.pypa.io/en/latest/references/keywords.html#keyword-namespace-packages for details.'),
+    ''
+)
+ignore_native_namespace_warning_5 = (re.compile(
+    r'ep.load\(\)\(self, ep.name, value\)'),
+    ''
+)
+
+
 def run_buildout(command):
     # Make sure we don't get .buildout
     os.environ['HOME'] = os.path.join(os.getcwd(), 'home')
     args = command.split()
-    import pkg_resources
     buildout = pkg_resources.load_entry_point(
         'zc.buildout', 'console_scripts', args[0])
     buildout(args[1:])
@@ -622,7 +660,11 @@ def run_from_process(target, *args, **kw):
     target(*args, **kw)
 
 def run_in_process(*args, **kwargs):
-    process = Process(target=run_from_process, args=args, kwargs=kwargs)
+    try:
+        ctx = multiprocessing.get_context('fork')
+        process = ctx.Process(target=run_from_process, args=args, kwargs=kwargs)
+    except AttributeError:
+        process = multiprocessing.Process(target=run_from_process, args=args, kwargs=kwargs)
     process.daemon = True
     process.start()
     process.join(99)
@@ -641,3 +683,41 @@ def run_buildout_in_process(command='buildout'):
         )
     command = ' '.join(command)
     run_in_process(run_buildout, command)
+
+
+def setup_coverage(path_to_coveragerc):
+    if 'RUN_COVERAGE' not in os.environ:
+        return
+    if not os.path.exists(path_to_coveragerc):
+        raise ValueError('coveragerc file %s does not exist.' % path_to_coveragerc)
+    os.environ['COVERAGE_PROCESS_START'] = path_to_coveragerc
+    rootdir = os.path.dirname(path_to_coveragerc)
+
+    def combine_report():
+        subprocess.call(
+            [
+                sys.executable, '-m', 'coverage', 'combine',
+            ],
+            cwd=rootdir,
+        )
+        subprocess.call(
+            [
+                sys.executable, '-m', 'coverage', 'report',
+            ],
+            cwd=rootdir,
+        )
+
+    if path_to_coveragerc:
+        try:
+            import coverage
+            print("Coverage configured with %s" % path_to_coveragerc)
+            if 'COVERAGE_REPORT' in os.environ:
+                import atexit
+                atexit.register(combine_report)
+            coverage.process_startup()
+        except ImportError:
+            print(
+                "You try to run coverage "
+                "but coverage is not installed in your environment."
+            )
+            sys.exit(1)
